import time as _time
import contextlib as _contextlib
from .core import find, prop_eq, uniq, concat, curry


@curry(2)
def p(label, v):
    print(label, v)
    return v


@_contextlib.contextmanager
def timed(label):
    now = _time.time()

    yield

    print(label, _time.time() - now)


def diff_collections(old, new):
    diff = {
        'added': [],
        'changed': [],
        'removed': [
            old_item for old_item in old
            if not find(prop_eq.c('id', old_item['id']), new)
        ],
    }

    for new_item in new:
        old_item = find(prop_eq.c('id', new_item['id']), old)

        if not old_item:
            diff['added'].append(new_item)
        elif new_item != old_item:
            diff['changed'].append({
                'from': old_item,
                'to': new_item,
            })

    return diff


def diff_dicts(a, b):
    diff = []

    for k in uniq(concat(a.keys(), b.keys())):
        if k not in a and k in b:
            diff.append(f'key "{k}" is new. New value: {b[k]}')
        elif k not in b and k in a:
            diff.append(f'key "{k}" was removed. Old value: {a[k]}')
        elif a[k] != b[k]:
            diff.append(
                f'key "{k}" was changed. Old value: {a[k]}; New value: {b[k]}'
            )

    return diff



import os, sys
from datetime import datetime
from .text import parse_bool


def now():
    return datetime.utcnow().isoformat() + 'Z'


def env(k, **kw):
    if 'default' in kw and k not in os.environ:
        return kw['default']

    v = os.environ[k]

    try:
        return int(v)
    except ValueError:
        pass

    try:
        return float(v)
    except ValueError:
        pass

    try:
        return parse_bool(v)
    except ValueError:
        pass

    return v


class Progress(object):
    width = 80

    def __init__(self, total):
        self.total = max(1, total)

    def show(self, i):
        j = round(i) / self.total
        spinner = ["# ", " #"][round(i) % 2]
        filled_w = self.width * j
        empty_w = self.width - filled_w
        fill = '=' * round(filled_w)
        space = ' ' * round(empty_w)

        sys.stdout.write('\r')
        sys.stdout.write(f"{spinner}[{fill}{space}] {int(100 * j)}%")
        sys.stdout.flush()

    def done(self):
        sys.stdout.write('\r')
        sys.stdout.write(' ' * (self.width + 8))
        sys.stdout.write('\r')
        sys.stdout.flush()



class Error(Exception):
    def __init__(self, code, message, meta={}):
        self.code = code
        self.message = message
        self.meta = meta

    def __repr__(self):
        return '<{} {}: {}>'.format(
            self.__class__.__name__,
            self.code,
            self.message
        )

    def has_code(self, code):
        return code in self.code

    def has_message(self, substring):
        return substring in self.message
